# Demonstration on how to apply the semi-extended source correction, in addition to point-source calibration,
# to a PACS SED spectrum of a semi-extended source with specExtendToPointCorrection task
# A semi-extended source is anything (spatial extent <~15" diameter) 
#
# - This works best for pointed chop nod observations, from which we will work on the rebinned cubes from Level 2
# - In fact it can also work on mapping observations, but that case you need to know which raster position, i.e. which 
#   cubes in "slicedFinalCube", have the source centred in the FoV, and you should select out only these slices to work on
#   More about selecting slices, etc, can be found in the spectroscopy Data Reduction Guide.
# - The script will also work on unchopped observations, for which you will take the final data from Level 2 (line scans)  
#   or Level 2.5 (range scans). However, note that the continuum uncertainty for unchopped observations is +/-20 Jy, and hence 
#   any (multiplicative) corrections computed for unchopped spectra will be even more uncertain. 
#
# This script demonstrates how to perform the E2P correction
# Additional material can be found in the package in which this script was bundled
# What is not provided in this script is how to create a model of a source that is not a Gaussian 
# (which is the demo model) or a Top Hat (which is the default model of the task)
#
# Created by E. Puga 23 April 2014 for the Spectrometer workshop
# Modified by K. Exter for general release Dec 2017
#




######################################################
# Directories, source name, model parameters, obsids # 
######################################################

scriptsdir  = '/Users/katrina/ownCloud/HSC_work/E2P/' # where the E2P scripts are kept
datapath   = '/Users/katrina/ownCloud/HSC_work/E2P/Data/' # where the observation and any other data are kept and can be written to
execfile(scriptsdir+"E2P_functions.py")
calTree=getCalTree()

## Chose the obsids of your source
# Here we are working with 1 obsid which covers half the blue and half of the red range of PACS for the source NGC 7027
obsids = [1342231960]
sourceName = 'NGC7027'

## Then work out the shape of the source and offset of the source from the centre of the field
# For this target we will use a model of a Gaussian (which is a simplication of the true source shape,
# but will serve for this demonstration)
model      = 'Gaussian' 
#model='TopHat' # to use the default top hat model of the E2P task instead

## Work out the fwhm of the source and its offset from the centre of the field
# For the Gaussian model, this diameter is the FWHM of the Gaussian; for the TopHat it is the diameter of the hat
diam, offRa, offDec = 12.0, 0.4, -0.7 # in arcsecs
# For the TopHat model you can also define a source position angle 
# (Clockwise position angle between the disc and the RA/Dec coordinate system (degrees). 
# 0.0 means that the X axis is aligned with the declination axis.) e.g.
#srcAng = 0.0

"""
How can one determine the offset? This depends on your knowledge of the source and we only give some 
general guidelines here. 

One way is to look at the cubes and find where the source peaks (this works best for a single-peaked source).
To do this one can display the source in the Cube Viewer in a layer where it is bright.
(One way to find that layer is to load the interpolated cubes into the Spectrum Explorer
 icube = obs.level2.blue.icube.product.get(0) # there is only one cube in the blue for this observation
then double-click on "icube" in the Variables panel and the SE will open in the Editor panel. 
Move the red line in the spectrum plot to where the spectrum is bright (e.g. an emission line) and note the layer 
value this is at. Ten display the cube with the Cube Viewer,  which you can set  to display that layer.
For this example, the layer 1181 is in the 88 micron line. 

From the Cube Viewer you can right-click and ask to "extract the current layer", which then appears 
in the Variables as "extractedLayer". (Double-click on this to display it.) 
Then use the source fitting task to find the centre of the image. (The Data Analysis Guide tells you how to 
use this task.) e.g. 
 parameters = sourceFitting(elongated=True, slope=False, image=extractedLayer, minX = 4, minY = 4, width=11, height=11)
(i.e. we are asking it to find the centre of the source in image subsection 4,4 to 15,15)
The RA and Dec are given in the parameters produced by sourceFitting. Double-click on it in Variables to view the parameters. 
For the example this is 21 07 01.603 42 14 09.3 

Next you need to compare this RA and Dec to that of the centre of the field, which you can get from the 
rebinned cube:
 rcube = obs.level2.blue.rcube.product.get(0)
 print MEAN(rcube.getRa()[:,2,2]), MEAN(rcube.getDec()[:,2,2])
here we have 316.756427056, 42.2361209611 in decimal or 21:7:1.542, 42:14:10.0, leading to an offset of approx. 
+0.37" in RA and -0.73" in Dec. 

Another way is to get an image of the source and use the PACS Spectral Footprint Viewer to drag and drop the 
rebinned cube on the image, to see where the source lies on the IFU footprint, i.e. where in which spaxel of the 
cube the photometer image shows that the source is centred.
The use of this viewer is explained in the PACS Data Reduction Guide for spectroscopy.

When determining the diameter of the source, remember it is not the measured diameter from the PACS images you 
want but rather its real diameter at the wavelength of the spectrum. Here we adopt a FWHM of 12" for the purpose 
of demonstration. 
(The PACS spectrometer and photometer beam sizes are given in the Quick Start Guides)

"""

## Define the source model 
## See E2P_functions.py. Here we are adopting a Circular Gaussian with offset wrt centre of central spaxel
## The fwhm is the fwhm of your Gaussian-like source, in arcsec 
source = MyGaussSource(diam, offRa, offDec)

## Model Visualization 
# The display is of the image here with a 0.5" pixel size and 109 pixels on a side) 
# (The appearance of the source will seem to change as you change these parameters, but this is just 
#   because of the displayable flux range)
Display(SimpleImage(image=source.getSourceModelImage(109, 0.5)))



############
# The data #
############


## Grab the existing blue and red rebinned cubes from the observation(s) and concatenate them
## Put as many or as few in here as you want to work on in one go
## What is in allCubes are slicedRebinnedCubes - that is, the HPD3DR[RB] product you can find in 
## Level 2 of an observation. Within in each slicedRebinnedCubes - called slicedFinalCubes here - can 
## be more than one actual rebinned cube, for example if several wavelength ranges were covered in the
## observation. Each of these cubes in a single slicedRebinnedCubes is called a "slice" later
allCubes=[]
for obsid in obsids:
   obs = getObservation(obsid, useHsa=1)
   for camera in ['blue','red']:
      slicedFinalCubes = obs.level2.getCamera(camera).rcube.product
      allCubes.append(slicedFinalCubes)
allCubes=concatenateSliced(allCubes)

## Plot of spectrum in central spaxel
x,y=2,2
pinit = plotCubes(allCubes,[],x=x,y=y)



############################################
# Point Source Correction & E2P Correction #
############################################
## The first step of the extended source correction is to extract the spectrum from the central 
## spaxel of the rebinned cubes and apply the point source correction. After that, the E2P - extended T- point - 
## correction is computed. It is this correction that requires the source model.

## Set up
segments = Int1d()
flags = Int1d()
fullSpec1 = Spectrum1d() # to contain the point source corrected result (no extended correction)
fullSpec9 = Spectrum1d()
fullSpec129 = Spectrum1d()
fullSpecCorr1 = Spectrum1d() # the full E2P corrected spectrum
fullSpecCorr9 = Spectrum1d()
fullSpecCorr129 = Spectrum1d()
fullCorrection1 = Spectrum1d() # the E2P correction to be applied to the PS spectrum 
fullCorrection9 = Spectrum1d()
fullCorrection129 = Spectrum1d()

verbose=0 # for extractCentralSpectrum (the verbose output are not so useful for semi-extended sources anyway)

## The parameters to specify for extractCentralSpectrum: these are taken from those used in the pipeline script 
## for range scans: see the pipeline script for line scans if you have that sort of data, or the URM entry for 
## the task 
smoothing = 'filter'
gaussianFilterWidth = 50
medianFilterWidth   = 15


## For each rebinned cube in allCubes, extract the spectrum of the central spaxel and apply the E2P correction
## This is done for each rebinned cube slice in each slicedRebinnedCubes in allCubes.
for slice in range(len(allCubes.refs)):
  # for each cube, create the point source corrected spectrum (in fact, 3 are produced by the task)
  rcube = allCubes.get(slice)
  band = rcube.meta['band'].value
  c1,c9,c129 = extractCentralSpectrum(rcube, smoothing=smoothing, width=gaussianFilterWidth, \
     preFilterWidth=medianFilterWidth,calTree=calTree, verbose=verbose)
  # Which one to use? This is up to the user (with guidance as given in the various documentation), here we produce all 3
  # If working with unchopped observations, use c1 (best) or c9 but not c129. For chop nod observations, c129 is usually the 
  # best, with the exception of very faint sources (c1).  
  if (model == 'Gaussian'):
     print "Gaussian model for slice",slice
     # in the case of a Gaussian, it is the FWHM
     source = MyGaussSource(diam, offRa, offDec)
     central3x3=False
     e2p1 = specExtendedToPointCorrection(c1,calTree=calTree,userSource=source,central3x3=central3x3, extra=True)   
     central3x3=True
     e2p9 = specExtendedToPointCorrection(c9,calTree=calTree,userSource=source,central3x3=central3x3, extra=True) 
     e2p129 = specExtendedToPointCorrection(c129,calTree=calTree,userSource=source,central3x3=central3x3, extra=True)     
  elif (model == 'TopHat'):
     # specExtendedToPointCorrection this assumes a disc top hat, ie it is the default 
     central3x3=False
     e2p1 = specExtendedToPointCorrection(c1,calTree=calTree,srcRadX=diam/2.,srcRadY=diam/2.,srcAngle=srcAng,raOffset=offRa,decOffset=offDec,central3x3=central3x3,extra=True)
     central3x3=True
     e2p9 = specExtendedToPointCorrection(c9,calTree=calTree,srcRadX=diam/2.,srcRadY=diam/2.,srcAngle=srcAng,raOffset=offRa,decOffset=offDec,central3x3=central3x3,extra=True)
     e2p129 = specExtendedToPointCorrection(c129,calTree=calTree,srcRadX=diam/2.,srcRadY=diam/2.,srcAngle=srcAng,raOffset=offRa,decOffset=offDec,central3x3=central3x3,extra=True)
  #
  c1e2pCorr = divide(ds1=c1, ds2=e2p1, overwrite=False)
  c9e2pCorr = divide(ds1=c9, ds2=e2p9, overwrite=False)
  c129e2pCorr = divide(ds1=c129, ds2=e2p129, overwrite=False)
  #
  # Concatenate all spectra from the slices to one spectrum1d, keeping them as separate segments
  # This way you can save the single product to FITS 
  #
  # First, just the point-source corrected spectra
  spec1 = c1.spectrum1d # c1 is a SimpleSpectrum and we convert it to Spectrum1d
  fullSpec1.concatenate(spec1)
  spec9 = c9.spectrum1d
  fullSpec9.concatenate(spec9)
  spec129 = c129.spectrum1d
  fullSpec129.concatenate(spec129)
  # Next the E2P corrected spectra
  spec1 = c1e2pCorr
  fullSpecCorr1.concatenate(spec1)
  spec9 = c9e2pCorr
  fullSpecCorr9.concatenate(spec9)
  spec129 = c129e2pCorr
  fullSpecCorr129.concatenate(spec129)
  # Now the correction curves
  corr = e2p1.spectrum1d
  fullCorrection1.concatenate(corr)
  corr = e2p9.spectrum1d
  fullCorrection9.concatenate(corr)
  corr = e2p129.spectrum1d
  fullCorrection129.concatenate(corr)
  # Fill in a segements and flag array (required inputs)
  segments.append(Int1d(spec1.flux.length(), slice))
  print "here",segments[-1]
  flg = Int1d(spec1.wave.length(), 0)
  flg=flagSet(band,1,flg,spec1.wave)
  flags.append(flg)
#

## Include the column segment and flags in the final Spectrum1d
fullSpec1.setSegment(segments)
fullSpec1.setFlag(flags)
fullSpec9.setSegment(segments)
fullSpec9.setFlag(flags)
fullSpec129.setSegment(segments)
fullSpec129.setFlag(flags)
fullSpecCorr1.setSegment(segments)
fullSpecCorr1.setFlag(flags)
fullSpecCorr9.setSegment(segments)
fullSpecCorr9.setFlag(flags)
fullSpecCorr129.setSegment(segments)
fullSpecCorr129.setFlag(flags)
fullCorrection1.setSegment(segments)
fullCorrection1.setFlag(flags)
fullCorrection9.setSegment(segments)
fullCorrection9.setFlag(flags)
fullCorrection129.setSegment(segments)
fullCorrection129.setFlag(flags)

## Save to disk
flnm = datapath+sourceName+"_E2P_FullyCorrected_c1.fits" # add whatever other information you want to the filename
simpleFitsWriter(fullSpecCorr1,flnm)
flnm = datapath+sourceName+"_E2P_FullyCorrected_c129.fits"
simpleFitsWriter(fullSpecCorr129,flnm)
flnm = datapath+sourceName+"_E2P_FullyCorrected_c9.fits"
simpleFitsWriter(fullSpecCorr9,flnm)

flnm = datapath+sourceName+"_E2P_CorrectionCurve_c1.fits"
simpleFitsWriter(fullCorrection1,flnm)
flnm = datapath+sourceName+"_E2P_CorrectionCurve_c129.fits"
simpleFitsWriter(fullCorrection129,flnm)
flnm = datapath+sourceName+"_E2P_CorrectionCure_c9.fits"
simpleFitsWriter(fullCorrection9,flnm)

flnm = datapath+sourceName+"_E2P_PSCorrectedOnly_c1.fits"
simpleFitsWriter(fullSpec1,flnm)
flnm = datapath+sourceName+"_E2P_PSCorrectedOnly_c129.fits" 
simpleFitsWriter(fullSpec129,flnm)
flnm = datapath+sourceName+"_E2P_PSCorrectedOnly_c9.fits" 
simpleFitsWriter(fullSpec9,flnm)

## Inspection
## First with the Spectrum Explorer and then with a plot
## (here plotting c129, you can do c1 and c9 also)
openVariable("fullSpec129", "Spectrum Explorer")

pC1=PlotXY()

npt=fullSpec129.getSegmentCount()
col=java.awt.Color.ORANGE
colShade=colorShade(col,npt+1) 
for i in range(npt):
   try:
      wave,flux,flag = fullSpec129.getSegment(i).wave, fullSpec129.getSegment(i).flux, fullSpec129.getSegment(i).flag
      idx= flag.where(flag == 0) 
      if i == npt/2:
         pC1.addLayer(LayerXY(wave[idx], flux[idx], line=1, color=colShade[i], inLegend=True, name="C129"))
      else:
         pC1.addLayer(LayerXY(wave[idx], flux[idx], line=1, color=colShade[i], inLegend=False))
   except:
      pass
col=java.awt.Color.RED
colShade=colorShade(col,npt+1) 
for i in range(npt):
   try:
      wave,flux,flag = fullSpecCorr129.getSegment(i).wave, fullSpecCorr129.getSegment(i).flux, fullSpecCorr129.getSegment(i).flag
      idx= flag.where(flag == 0) 
      if i == npt/2:
         pC1.addLayer(LayerXY(wave[idx], flux[idx], line=1, color=colShade[i], inLegend=True, name="C129/e2p129"))
      else:
         pC1.addLayer(LayerXY(wave[idx], flux[idx], line=1, color=colShade[i], inLegend=False))
   except:
      pass
pC1.legend.visible=1
pC1.titleText=sourceName
pC1.setXtitle("Wavelength ($\micro$m)")
pC1.setYtitle("Observed Flux (Jy)")
pC1.xaxis.tick.gridLines = 1
pC1.yaxis.tick.gridLines = 1
pC1.xaxis.range=[50, 192] 
pC1.setPlotSize(8, 3.0)
c = 299792458.0 # m/s
xaux = ReciprocalAuxAxis(1.0e-3*c)
pC1.getXaxis().removeAuxAxis(0)
pC1.getXaxis().addAuxAxis(xaux)
xaux.setTitleText("Frequency (GHz)")
pC1.xaxis.type=Axis.LOG
pC1.yaxis.type=Axis.LOG
pC1.xaxis.range=[50, 192] 
pC1.xaxis.range=[0, 200] 

## Display corrections E2P1 and E2P9 in plot
pCor=PlotXY()
npt=fullSpec1.getSegmentCount()
for i in range(npt):
   try:
      wave1,flux1,flag1 = fullCorrection1.getSegment(i).wave, fullCorrection1.getSegment(i).flux, fullCorrection1.getSegment(i).flag
      wave9,flux9,flag9 = fullCorrection9.getSegment(i).wave, fullCorrection9.getSegment(i).flux, fullCorrection9.getSegment(i).flag
      idx= flag1.where(flag1 == 0) 
      if i == 0:
         pCor.addLayer(LayerXY(wave1[idx], flux1[idx], line=1, color=java.awt.Color.BLUE, inLegend=True, name="e2p1"))
         pCor.addLayer(LayerXY(wave9[idx], flux9[idx], line=1, color=java.awt.Color.RED, inLegend=True, name="e2p9"))
      else:
         pCor.addLayer(LayerXY(wave1[idx], flux1[idx], line=1, color=java.awt.Color.BLUE, inLegend=False))     
         pCor.addLayer(LayerXY(wave9[idx], flux9[idx], line=1, color=java.awt.Color.RED, inLegend=False))
   except:
      pass
   pCor.addLayer(LayerXY(wave1[idx], flux1[idx]))
pCor.legend.visible=1
pCor.titleText='Source: %s, diam: %.1f, Offset: %0.2f, %0.2f \"' % (sourceName, diam, offRa, offDec)
pCor.setXtitle("Wavelength ($\micro$m)")
pCor.setYtitle("Factor")
pCor.xaxis.tick.gridLines = 1
pCor.yaxis.tick.gridLines = 1
pCor.xaxis.range=[50, 192]
pCor.setPlotSize(8, 3.0)

